import click
import subprocess as sp

from higlass_manage.common import hg_name_to_container_name


@click.command()
@click.option(
    "--hg-name",
    default="default",
    help="The name of the higlass container to import this file to",
)
def superuser(hg_name):
    """
    Create a superuser in the container

    Parameters
    ----------
    hg_name: string
        The name of the container to create a superuser on
    """
    container_name = hg_name_to_container_name(hg_name)
    sp.run(
        [
            "docker",
            "exec",
            "-it",
            container_name,
            "python",
            "higlass-server/manage.py",
            "createsuperuser",
        ]
    )



import docker
import hashlib
import os
import os.path as op
import slugid
import sys

CONTAINER_PREFIX = "higlass-manage-container"
NETWORK_PREFIX = "higlass-manage-network"
REDIS_PREFIX = "higlass-manage-redis"
REDIS_CONF = "/usr/local/etc/redis/redis.conf"
SQLITEDB = "db.sqlite3"


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def hg_name_to_container_name(hg_name):
    return "{}-{}".format(CONTAINER_PREFIX, hg_name)


def hg_name_to_network_name(hg_name):
    return "{}-{}".format(NETWORK_PREFIX, hg_name)


def hg_name_to_redis_name(hg_name):
    return "{}-{}".format(REDIS_PREFIX, hg_name)


def get_port(hg_name):
    client = docker.from_env()

    container_name = hg_name_to_container_name(hg_name)
    config = client.api.inspect_container(container_name)
    port = config["HostConfig"]["PortBindings"]["80/tcp"][0]["HostPort"]

    return port


def get_site_url(hg_name, _SITE_URL="SITE_URL"):
    """
    get SITE_URL for a given container
    using container docker-config, assuming
    there is not more than one SITE_URL in
    container's env.

    Yields "localhost" when no SITE_URL entries
    detected.
    """

    client = docker.from_env()

    container_name = hg_name_to_container_name(hg_name)
    config = client.api.inspect_container(container_name)

    site_url_entries = [s for s in config["Config"]["Env"] if _SITE_URL in s]
    # if there is no SITE_URL entry yield "localhost"
    if not site_url_entries:
        return "http://localhost"
    # otherwise there has to be only one SITE_URL entry in the Env:
    elif len(site_url_entries) > 1:
        raise ValueError(
            "There are multiple SITE_URL entry in {} env".format(container_name)
        )
    else:
        (site_url,) = site_url_entries
        # parse "SITE_URL=http://hostname":
        _, site_url = site_url.split("=")
        return site_url


def fill_filetype_and_datatype(filename, filetype, datatype):
    """
    If no filetype or datatype are provided, add them
    based on the given filename.

    Parameters:
    ----------
    filename: str
        The name of the file
    filetype: str
        The type of the file (can be None)
    datatype: str
        The datatype for the data in the file (can be None)

    Returns:
    --------
    (filetype, datatype): (str, str)
        Filled in filetype and datatype based on the given filename
    """
    if filetype is None:
        # no filetype provided, try a few common filetypes
        filetype = infer_filetype(filename)
        print("Inferred filetype:", filetype)

        if filetype is None:
            recommended_filetype = recommend_filetype(filename)

            print(
                "Unknown filetype, please specify using the --filetype option",
                file=sys.stderr,
            )
            if recommended_filetype is not None:
                print(
                    "Based on the filename, you may want to try the filetype: {}".format(
                        recommended_filetype
                    )
                )

            return (None, None)

    if datatype is None:
        datatype = infer_datatype(filetype)
        print("Inferred datatype:", datatype)

        if datatype is None:
            recommended_datatype = recommend_datatype(filetype)
            print(
                "Unknown datatype, please specify using the --datatype option",
                file=sys.stderr,
            )
            if recommended_datatype is not None:
                print(
                    "Based on the filetype, you may want to try the datatype: {}".format(
                        recommended_datatype
                    )
                )

    return (filetype, datatype)


def recommend_filetype(filename):
    ext = op.splitext(filename)
    if op.splitext(filename)[1] == ".bed":
        return "bedfile"
    if op.splitext(filename)[1] == ".bedpe":
        return "bedpe"


def recommend_datatype(filetype):
    if filetype == "bedfile":
        return "bedlike"


def datatype_to_tracktype(datatype):
    if datatype == "matrix":
        return ("heatmap", "center")
    elif datatype == "vector":
        return ("horizontal-bar", "top")
    elif datatype == "gene-annotations":
        return ("horizontal-gene-annotations", "top")
    elif datatype == "chromsizes":
        return ("horizontal-chromosome-labels", "top")
    elif datatype == "2d-rectangle-domains":
        return ("2d-rectangle-domains", "center")
    elif datatype == "bedlike":
        return ("bedlike", "top")

    return (None, None)


def infer_filetype(filename):
    _, ext = op.splitext(filename)

    if ext.lower() == ".bw" or ext.lower() == ".bigwig":
        return "bigwig"
    elif ext.lower() == ".mcool" or ext.lower() == ".cool":
        return "cooler"
    elif ext.lower() == ".htime":
        return "time-interval-json"
    elif ext.lower() == ".hitile":
        return "hitile"
    elif ext.lower() == ".beddb":
        return "beddb"

    return None


def infer_datatype(filetype):
    if filetype == "cooler":
        return "matrix"
    if filetype == "bigwig":
        return "vector"
    if filetype == "time-interval-json":
        return "time-interval"
    if filetype == "hitile":
        return "vector"
    if filetype == "beddb":
        return "bedlike"


def import_file(
    hg_name, filepath, filetype, datatype, assembly, name, uid, no_upload, project_name
):
    # get this container's temporary directory
    if not no_upload:
        temp_dir = get_temp_dir(hg_name)
        if not op.exists(temp_dir):
            os.makedirs(temp_dir)

        filename = op.split(filepath)[1]
        to_import_path = op.join(temp_dir, filename)

        if to_import_path != filepath:
            # if this file already exists in the temporary dir
            # remove it
            if op.exists(to_import_path):
                print("Removing existing file in temporary dir:", to_import_path)
                os.remove(to_import_path)

            os.link(filepath, to_import_path)
    else:
        filename = filepath

    coordSystem = "--coordSystem {}".format(assembly) if assembly is not None else ""
    name_text = '--name "{}"'.format(name) if name is not None else ""
    project_name_text = (
        '--project-name "{}"'.format(project_name) if project_name is not None else ""
    )

    print("name_text: {}".format(name_text))

    client = docker.from_env()
    print("hg_name:", hg_name)
    container_name = hg_name_to_container_name(hg_name)
    container = client.containers.get(container_name)

    if no_upload:
        command = (
            "python higlass-server/manage.py ingest_tileset --filename"
            + " {}".format(filename.replace(" ", "\ "))
            + " --filetype {} --datatype {} {} {} {} --no-upload".format(
                filetype, datatype, name_text, project_name_text, coordSystem
            )
        )
    else:
        command = (
            "python higlass-server/manage.py ingest_tileset --filename"
            + " /tmp/{}".format(filename.replace(" ", "\ "))
            + " --filetype {} --datatype {} {} {} {}".format(
                filetype, datatype, name_text, project_name_text, coordSystem
            )
        )

    if uid is not None:
        command += " --uid {}".format(uid)
    else:
        uid = slugid.nice()
        command += " --uid {}".format(uid)

    print("command:", command)

    (exit_code, output) = container.exec_run(command)

    if exit_code != 0:
        print("ERROR:", output.decode("utf8"), file=sys.stderr)
        return None

    return uid


def get_temp_dir(hg_name):
    client = docker.from_env()
    container_name = hg_name_to_container_name(hg_name)
    config = client.api.inspect_container(container_name)

    print("state", config["State"]["Running"])

    if config["State"]["Running"] != True:
        raise HiGlassNotRunningException()

    for mount in config["Mounts"]:
        if mount["Destination"] == "/tmp":
            return mount["Source"]


def get_data_dir(hg_name):
    client = docker.from_env()
    container_name = hg_name_to_container_name(hg_name)
    config = client.api.inspect_container(container_name)

    for mount in config["Mounts"]:
        if mount["Destination"] == "/data":
            return mount["Source"]


class HiGlassNotRunningException(Exception):
    pass



import click
import docker
import os
import sys
import webbrowser

from higlass_manage.common import get_port


@click.command()
@click.argument("names", nargs=-1)
def browse(names):
    """
    Launch a web browser for a running instance
    """
    client = docker.from_env()

    if len(names) == 0:
        names = ("default",)

    try:
        port = get_port(names[0])
    except docker.errors.NotFound:
        print(
            "Error: higlass instance not found. Have you tried starting it using 'higlass-manage start'?",
            file=sys.stderr,
        )
        return

    # make sure this test passes on Travis CI and doesn't try to open
    # a terminal-based browser which doesn't return
    if not os.environ.get("HAS_JOSH_K_SEAL_OF_APPROVAL"):
        webbrowser.open("http://localhost:{port}/app/".format(port=port))



import click
import docker
import json
import os
import os.path as op
import requests
import slugid
import sys
import time

from higlass_manage.common import (
    CONTAINER_PREFIX,
    NETWORK_PREFIX,
    REDIS_PREFIX,
    REDIS_CONF,
)


@click.command()
@click.option(
    "-t",
    "--temp-dir",
    default="/tmp/higlass-docker",
    help="The temp directory to use",
    type=str,
)
@click.option(
    "-d",
    "--data-dir",
    default="~/hg-data",
    help="The higlass data directory to use",
    type=str,
)
@click.option(
    "-v",
    "--version",
    default="latest",
    help='The version of the Docker container to use (or "local" for images built locally using higlass-docker)',
    type=str,
)
@click.option(
    "-p",
    "--port",
    default=8989,
    help="The port that the HiGlass instance should run on",
    type=str,
)
@click.option(
    "-n",
    "--hg-name",
    default="default",
    help="The name for this higlass instance",
    type=str,
)
@click.option(
    "-s",
    "--site-url",
    default=None,
    help="When creating an external-facing instance, enter its IP or hostname using this parameter",
    type=str,
)
@click.option(
    "-m",
    "--media-dir",
    default=None,
    help="Use a specific media directory for uploaded files.",
    type=str,
)
@click.option(
    "--public-data/--no-public-data",
    default=True,
    help="Include or exclude public data in the list of available tilesets",
)
@click.option(
    "--default-track-options",
    default=None,
    help="Specify a json file containing default track options",
)
@click.option(
    "--workers",
    default=None,
    help="Specify a custom number of workers for the uWSGI application server",
)
@click.option(
    "--use-redis",
    is_flag=True,
    help="Initialize a Redis-based caching service and bind higlass instance to it",
)
@click.option(
    "--redis-dir",
    default="~/redis-data",
    help="Use a specific directory for Redis files",
    type=str,
)
@click.option(
    "--hg-repository",
    default="higlass/higlass-docker",
    help="The Docker repository to use for the HiGlass image",
    type=str,
)
@click.option(
    "--redis-repository",
    default="redis",
    help="The Docker repository to use for the Redis image",
    type=str,
)
@click.option(
    "--redis-tag",
    default="5.0.3-alpine",
    help="The Docker tag to use for the Redis image repository",
    type=str,
)
@click.option(
    "--redis-port", default=6379, help="The port to use for the Redis image", type=int
)
def start(
    temp_dir,
    data_dir,
    version,
    port,
    hg_name,
    site_url,
    media_dir,
    public_data,
    default_track_options,
    workers,
    use_redis,
    redis_dir,
    hg_repository,
    redis_repository,
    redis_tag,
    redis_port,
):
    _start(
        temp_dir,
        data_dir,
        version,
        port,
        hg_name,
        site_url,
        media_dir,
        public_data,
        default_track_options,
        workers,
        use_redis,
        redis_dir,
        hg_repository,
        redis_repository,
        redis_tag,
        redis_port,
    )


def _start(
    temp_dir="/tmp/higlass-docker",
    data_dir="~/hg-data",
    version="latest",
    port=8989,
    hg_name="default",
    site_url=None,
    media_dir=None,
    public_data=True,
    default_track_options=None,
    workers=None,
    use_redis=False,
    redis_dir="~/redis-data",
    hg_repository="higlass/higlass-docker",
    redis_repository="redis",
    redis_tag="5.0.3-alpine",
    redis_port=6379,
):
    """
    Start a HiGlass instance
    """
    hg_container_name = "{}-{}".format(CONTAINER_PREFIX, hg_name)

    client = docker.from_env()

    try:
        hg_container = client.containers.get(hg_container_name)

        sys.stderr.write("Stopping previously running container\n")
        hg_container.stop()
        hg_container.remove()
    except docker.errors.NotFound:
        # container isn't running so no need to stop it
        pass
    except requests.exceptions.ConnectionError:
        sys.stderr.write(
            "Error connecting to the Docker daemon, make sure it is started and you are logged in.\n"
        )
        return

    if use_redis:
        network_name = "{}-{}".format(NETWORK_PREFIX, hg_name)
        redis_name = "{}-{}".format(REDIS_PREFIX, hg_name)

        # set up a bridge network for Redis and higlass containers to share
        try:
            network_list = client.networks.list(names=[network_name])
            if network_list:
                network = client.networks.get(network_name)
                sys.stderr.write(
                    "Attempting to remove existing Docker network instance\n"
                )
                network.remove()
        except docker.errors.APIError:
            sys.stderr.write(
                "Error: Could not access Docker network list to remove existing network.\n"
            )
            sys.exit(-1)

        try:
            # https://docker-py.readthedocs.io/en/stable/networks.html
            network = client.networks.create(network_name, driver="bridge")
        except docker.errors.APIError as err:
            sys.stderr.write(
                "Error: Could not access Docker network ({}).\n".format(err)
            )
            sys.exit(-1)

        # clear up any running Redis container
        try:
            redis_container = client.containers.get(redis_name)
            sys.stderr.write("Stopping previously running Redis container\n")
            redis_container.stop()
            redis_container.remove()
        except docker.errors.NotFound:
            pass
        except requests.exceptions.ConnectionError:
            sys.stderr.write(
                "Error: Error connecting to the Docker daemon, make sure it is started and you are logged in.\n"
            )
            sys.exit(-1)

        # pull Redis image
        sys.stderr.write("Pulling {}:{}\n".format(redis_repository, redis_tag))
        sys.stderr.flush()
        redis_image = client.images.pull(redis_repository, tag=redis_tag)
        sys.stderr.write("done\n")
        sys.stderr.flush()

        # set up Redis container settings and environment
        redis_dir = op.expanduser(redis_dir)

        if not op.exists(redis_dir):
            os.makedirs(redis_dir)

        redis_conf = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "redis", "redis.conf"
        )
        if not os.path.exists(redis_conf):
            sys.stderr.write(
                "Error: Could not locate Redis configuration file [{}]\n".format(
                    redis_conf
                )
            )
            sys.exit(-1)

        redis_volumes = {
            redis_dir: {"bind": "/data", "mode": "rw"},
            redis_conf: {"bind": REDIS_CONF, "mode": "rw"},
        }

        redis_command = "redis-server {}".format(REDIS_CONF)

        try:
            # run Redis container
            redis_container = client.containers.run(
                redis_image,
                redis_command,
                name=redis_name,
                network=network_name,
                volumes=redis_volumes,
                detach=True,
            )
        except docker.errors.ContainerError as err:
            sys.stderr.write(
                "Error: Redis container could not be started\n{}\n".format(err)
            )
            sys.exit(-1)
        except docker.errors.ImageNotFound as err:
            sys.stderr.write(
                "Error: Redis container image could not be found\n{}\n".format(err)
            )
            sys.exit(-1)
        except docker.errors.APIError as err:
            sys.stderr.write(
                "Error: Redis container server ran into a fatal error\n{}\n".format(err)
            )
            sys.exit(-1)

    if version == "local":
        hg_image = client.images.get("image-default")
    else:
        sys.stderr.write("Pulling latest image... \n")
        sys.stderr.flush()
        hg_image = client.images.pull(hg_repository, version)
        sys.stderr.write("done\n")
        sys.stderr.flush()

    data_dir = op.expanduser(data_dir)
    temp_dir = op.expanduser(temp_dir)

    if not op.exists(temp_dir):
        os.makedirs(temp_dir)
    if not op.exists(data_dir):
        os.makedirs(data_dir)

    hg_environment = {}

    if site_url is not None:
        hg_environment["SITE_URL"] = site_url

    if workers is not None:
        hg_environment["WORKERS"] = workers

    sys.stderr.write("Data directory: {}\n".format(data_dir))
    sys.stderr.write("Temp directory: ()\n".format(temp_dir))

    hg_version_addition = "" if version is None else ":{}".format(version)

    sys.stderr.write("Starting... {} {}\n".format(hg_name, port))
    hg_volumes = {
        temp_dir: {"bind": "/tmp", "mode": "rw"},
        data_dir: {"bind": "/data", "mode": "rw"},
    }

    if media_dir:
        hg_volumes[media_dir] = {"bind": "/media", "mode": "rw"}
        hg_environment["HIGLASS_MEDIA_ROOT"] = "/media"

    if not use_redis:
        hg_container = client.containers.run(
            hg_image,
            ports={80: port},
            volumes=hg_volumes,
            name=hg_container_name,
            environment=hg_environment,
            detach=True,
        )
    else:
        # add some environment variables to the higlass container
        hg_environment["REDIS_HOST"] = redis_name
        hg_environment["REDIS_PORT"] = redis_port
        # run the higlass container on the shared network with the Redis container
        hg_container = client.containers.run(
            hg_image,
            network=network_name,
            ports={80: port},
            volumes=hg_volumes,
            name=hg_container_name,
            environment=hg_environment,
            publish_all_ports=True,
            detach=True,
        )

    sys.stderr.write("Docker started: {}\n".format(hg_container_name))

    started = False
    counter = 1
    while not started:
        try:
            sys.stderr.write("sending request {}\n".format(counter))
            counter += 1
            req = requests.get(
                "http://localhost:{}/api/v1/viewconfs/?d=default".format(port),
                timeout=5,
            )
            # sys.stderr.write("request returned {} {}\n".format(req.status_code, req.content))

            if req.status_code != 200:
                sys.stderr.write(
                    "Non 200 status code returned ({}), waiting...\n".format(
                        req.status_code
                    )
                )
                time.sleep(0.5)
            else:
                started = True
        except requests.exceptions.ConnectionError:
            sys.stderr.write("Waiting to start (tilesets)...\n")
            time.sleep(0.5)
        except requests.exceptions.ReadTimout:
            sys.stderr.write("Request timed out\n")
            time.sleep(0.5)

    sys.stderr.write("public_data: {}\n".format(public_data))

    if not public_data or default_track_options is not None:
        # we're going to be changing the higlass js file so first we copy it to a location
        # with a new hash
        new_hash = slugid.nice()

        sed_command = """bash -c 'cp higlass-app/static/js/main.*.chunk.js higlass-app/static/js/main.{}.chunk.js'""".format(
            new_hash
        )

        ret = hg_container.exec_run(sed_command)

    if not public_data:
        config = json.loads(req.content.decode("utf-8"))
        config["trackSourceServers"] = ["/api/v1"]
        started = True
        # sys.stderr.write('config {}\n'.format(json.dumps(config, indent=2)))
        config = {"uid": "default_local", "viewconf": config}

        ret = hg_container.exec_run(
            """python higlass-server/manage.py shell --command="import tilesets.models as tm; o = tm.ViewConf.objects.get(uuid='default_local'); o.delete();" """
        )
        ret = requests.post(
            "http://localhost:{}/api/v1/viewconfs/".format(port), json=config
        )
        sys.stderr.write("ret: {}\n".format(ret.content))
        # ret = container.exec_run('echo "import tilesets.models as tm; tm.ViewConf.get(uuid={}default{}).delete()" | python higlass-server/manage.py shell'.format("'", "'"), tty=True)
        ret = hg_container.exec_run(
            """bash -c 'sed -i '"'"'s/"default"/"default_local"/g'"'"' higlass-app/static/js/main.*.chunk.js'""".format(
                new_hash
            )
        )
        sys.stderr.write("ret: {}\n".format(ret))

    if default_track_options is not None:
        with open(default_track_options, "r") as f:
            default_options_json = json.load(f)

            sed_command = """bash -c 'sed -i '"'"'s/assign({{}},this.props.options/assign({{defaultOptions: {} }},this.props.options/g'"'"' """.format(
                json.dumps(default_options_json)
            )
            sed_command += " higlass-app/static/js/main.*.chunk.js'"
            # sys.stderr.write("sed_command: {}\n".format(sed_command))

            ret = hg_container.exec_run(sed_command)

    # sys.stderr.write("ret: {}\n".format(ret))

    sed_command = (
        """bash -c 'sed -i '"'"'s/main.*.chunk.js/main.invalid.chunk.js/g'"'"' """
    )
    sed_command += " higlass-app/precache-manifest.*.js'"

    ret = hg_container.exec_run(sed_command)
    # sys.stderr.write("ret: {}\n".format(ret))

    sed_command = """bash -c 'sed -i '"'"'s/index.html/index_invalidated_by_higlass_manage/g'"'"' """
    sed_command += " higlass-app/precache-manifest.*.js'"

    ret = hg_container.exec_run(sed_command)
    # sys.stderr.write("ret: {}\n".format(ret))
    sys.stderr.write("Replaced js file\n")

    sys.stderr.write("Started\n")



import sys
import click
import docker
import json
import requests

from higlass_manage.common import get_port
from higlass_manage.common import CONTAINER_PREFIX, REDIS_PREFIX


@click.command()
@click.option(
    "--hg-name",
    default="default",
    help="The name of the higlass container to import this file to",
)
def tilesets(hg_name):
    """
    List the datasets in an instance
    """
    port = get_port(hg_name)

    # use a really high limit to avoid paging
    url = "http://localhost:{}/api/v1/tilesets/?limit=10000".format(port)
    ret = requests.get(url)

    if ret.status_code != 200:
        sys.stderr.write("Error retrieving tilesets: {}".format(ret))
        return

    j = json.loads(ret.content.decode("utf8"))
    for result in j["results"]:
        sys.stdout.write(
            "{}\n".format(
                " | ".join(
                    [
                        result["uuid"],
                        result["filetype"],
                        result["datatype"],
                        result["coordSystem"],
                        result["name"],
                    ]
                )
            )
        )


@click.command()
def instances():
    """
    List running instances
    """
    client = docker.from_env()

    for container in client.containers.list():
        name = container.name
        if name.find(CONTAINER_PREFIX) == 0:
            hm_name = name[len(CONTAINER_PREFIX) + 1 :]
            config = client.api.inspect_container(container.name)
            directories = " ".join(
                [
                    "{}:{}".format(m["Source"], m["Destination"])
                    for m in config["Mounts"]
                ]
            )
            port = config["HostConfig"]["PortBindings"]["80/tcp"][0]["HostPort"]
            sys.stdout.write("higlass\t{}\t{}\t{}\n".format(hm_name, directories, port))
        if name.find(REDIS_PREFIX) == 0:
            redis_name = name[len(REDIS_PREFIX) + 1 :]
            redis_config = client.api.inspect_container(container.name)
            redis_directories = " ".join(
                [
                    "{}:{}".format(m["Source"], m["Destination"])
                    for m in redis_config["Mounts"]
                ]
            )
            sys.stdout.write("redis\t{}\t{}\t.\n".format(redis_name, redis_directories))



import sys
import click
import docker

from .common import CONTAINER_PREFIX, NETWORK_PREFIX, REDIS_PREFIX


@click.command()
@click.argument("names", nargs=-1)
@click.option(
    "--remove-container/--dont-remove-container",
    default=True,
    show_default=True,
    help="Remove stopped higlass container",
)
@click.option(
    "--stop-redis/--dont-stop-redis",
    default=True,
    show_default=True,
    help="Stop and remove redis container"
    " associated with a given higlass"
    " instance.",
)
@click.option(
    "--remove-network-bridge/--dont-remove-network-bridge",
    default=True,
    show_default=True,
    help="Remove network bridge associated with a given higlass instance.",
)
def stop(
    names, remove_container, stop_redis, remove_network_bridge,
):
    _stop(
        names, remove_container, stop_redis, remove_network_bridge,
    )


def _stop(
    names, remove_container=True, stop_redis=True, remove_network_bridge=True,
):
    """
    Stop a running higlass instance along with the
    associated redis container and network bridges.

    The script attemps to stop and remove all of the
    containers/networks associated with a given higlass
    name.
    """
    client = docker.from_env()

    if len(names) == 0:
        names = ("default",)

    for name in names:
        # higlass container
        hm_name = "{}-{}".format(CONTAINER_PREFIX, name)
        try:
            client.containers.get(hm_name).stop()
            if remove_container:
                client.containers.get(hm_name).remove()
        except docker.errors.NotFound as ex:
            sys.stderr.write("Instance not running: {}\n".format(name))

        # redis container
        if stop_redis:
            redis_name = "{}-{}".format(REDIS_PREFIX, name)
            try:
                client.containers.get(redis_name).stop()
                client.containers.get(redis_name).remove()
            except docker.errors.NotFound:
                sys.stderr.write(
                    "No Redis instances found at {}; skipping...\n".format(redis_name)
                )

        # bridge network
        if remove_network_bridge:
            network_name = "{}-{}".format(NETWORK_PREFIX, name)
            try:
                network_list = client.networks.list(names=[network_name])
                if network_list:
                    network = client.networks.get(network_name)
                    network.remove()
            except docker.errors.NotFound:
                sys.stderr.write(
                    "No bridge network found at {}; skipping...\n".format(network_name)
                )



import click
import json
import ntpath
import os
import os.path as op
import requests
import sys
import webbrowser

from higlass_manage.common import fill_filetype_and_datatype
from higlass_manage.common import get_port
from higlass_manage.common import get_data_dir
from higlass_manage.common import get_temp_dir
from higlass_manage.common import md5
from higlass_manage.common import datatype_to_tracktype
from higlass_manage.start import _start
from higlass_manage.ingest import _ingest


@click.command()
@click.argument("filename", nargs=1)
@click.option(
    "-n",
    "--hg-name",
    default="default",
    help="The name for this higlass instance",
    type=str,
)
@click.option(
    "--filetype", default=None, help="The type of file to ingest (e.g. cooler)"
)
@click.option(
    "--datatype", default=None, help="The data type of in the input file (e.g. matrix)"
)
@click.option("--tracktype", default=None, help="The track type used to view this file")
@click.option(
    "--position", default=None, help="The position in the view to place this track"
)
@click.option(
    "--public-data/--no-public-data",
    default=True,
    help="Include or exclude public data in the list of available tilesets",
)
@click.option(
    "--assembly", default=None, help="The assembly that this data is mapped to"
)
@click.option(
    "--chromsizes-filename",
    default=None,
    help="A set of chromosome sizes to use for bed and bedpe files",
)
def view(
    filename,
    hg_name,
    filetype,
    datatype,
    tracktype,
    position,
    public_data,
    assembly,
    chromsizes_filename,
):
    """
    View a file in higlass.

    The user can specify an instance to view it in. If one is
    not specified the default will be used. If the default isn't
    running, it will be started.

    Parameters:
    -----------
    hg_name: string
        The name of the higlass instance
    filename: string
        The name of the file to view
    """
    try:
        temp_dir = get_temp_dir(hg_name)
        print("temp_dir:", temp_dir)
    except Exception:
        _start(hg_name=hg_name)

    # check if we have a running instance
    # if not, start one

    # get a list of the available tilesets
    # check if any match the filename of this file
    # if the filenames match, check if the checksums match
    port = get_port(hg_name)
    uuid = None

    # guess filetype and datatype if they're None
    (filetype, inferred_datatype) = fill_filetype_and_datatype(
        filename, filetype, datatype
    )

    if filetype is None or inferred_datatype is None:
        print(
            "Couldn't infer filetype or datatype ({}, {}),".format(
                filetype, inferred_datatype
            ),
            "please specify them using the command line options",
            file=sys.stderr,
        )
        return

    try:
        MAX_TILESETS = 100000
        req = requests.get(
            "http://localhost:{}/api/v1/tilesets/?limit={}".format(port, MAX_TILESETS),
            timeout=10,
        )

        tilesets = json.loads(req.content)

        for tileset in tilesets["results"]:
            import_filename = op.splitext(ntpath.basename(filename))[0]
            tileset_filename = ntpath.basename(tileset["datafile"])

            subpath_index = tileset["datafile"].find("/tilesets/")
            subpath = tileset["datafile"][subpath_index + len("/tilesets/") :]

            data_dir = get_data_dir(hg_name)
            tileset_path = op.join(data_dir, subpath)

            # print("import_filename", import_filename)
            # print("tileset_filename", tileset_filename)

            if tileset_filename.find(import_filename) >= 0:
                # same filenames, make sure they're actually the same file
                # by comparing checksums
                checksum1 = md5(tileset_path)
                checksum2 = md5(filename)

                if checksum1 == checksum2:
                    uuid = tileset["uuid"]
                    break
    except requests.exceptions.ConnectionError:
        print("Error getting a list of existing tilesets", file=sys.stderr)
        return

    if uuid is None:
        # we haven't found a matching tileset so we need to ingest this one
        uuid = _ingest(
            filename,
            hg_name,
            filetype,
            datatype,
            assembly=assembly,
            chromsizes_filename=chromsizes_filename,
        )

    if uuid is None:
        # couldn't ingest the file
        return

    from higlass.client import Track, View, ViewConf

    if datatype is None:
        datatype = inferred_datatype

    if tracktype is None and position is None:
        (tracktype, position) = datatype_to_tracktype(datatype)

        if tracktype is None:
            print("ERROR: Unknown track type for the given datatype:", datatype)
            return

    view = View(
        [
            Track(
                track_type=tracktype,
                position=position,
                tileset_uuid=uuid,
                server="http://localhost:{}/api/v1/".format(port),
                height=200,
            ),
        ]
    )

    viewconf = ViewConf([view])

    conf = viewconf.to_dict()

    conf["trackSourceServers"] = []
    conf["trackSourceServers"] += ["http://localhost:{}/api/v1/".format(port)]

    if public_data:
        conf["trackSourceServers"] += ["http://higlass.io/api/v1/"]

    # uplaod the viewconf
    res = requests.post(
        "http://localhost:{}/api/v1/viewconfs/".format(port), json={"viewconf": conf}
    )

    if res.status_code != 200:
        print("Error posting viewconf:", res.status, res.content)
        return

    uid = json.loads(res.content)["uid"]

    # make sure this test passes on Travis CI and doesn't try to open
    # a terminal-based browser which doesn't return
    if not os.environ.get("HAS_JOSH_K_SEAL_OF_APPROVAL"):
        webbrowser.open(
            "http://localhost:{port}/app/?config={uid}".format(port=port, uid=uid)
        )



import sys
import click
import os.path as op
import sqlite3
import shutil
import docker

from .common import (
    get_data_dir,
    get_site_url,
    get_port,
    hg_name_to_container_name,
    SQLITEDB,
)

from .stop import _stop


@click.command()
@click.option(
    "--old-hg-name",
    help="The name of the running higlass container that needs to be updated.",
    required=False,
)
@click.option(
    "--old-site-url",
    help="site-url at the old location."
    " Provide this when higlass container"
    " to be updated is not running.",
    required=False,
)
@click.option(
    "--old-port",
    help="port at the old location."
    " Provide this when higlass container"
    " to be updated is not running.",
    required=False,
    default="80",
    type=str,
)
@click.option(
    "--old-data-dir",
    help="data directory of the higlass"
    " that is to be updated (usually 'hg-data')."
    " Provide this when higlass container"
    " is not running.",
    required=False,
)
@click.option(
    "--new-site-url",
    default="http://localhost",
    help="site-url at the new location.",
    required=True,
)
@click.option(
    "--new-port",
    help="port at the new location",
    required=False,
    default="80",
    type=str,
)
@click.option(
    "--db-backup-name",
    help="name of the database (db) backup file."
    " db backup will be stored in the data directory"
    " provided explicitly or inferred from the running"
    " container.",
    required=False,
    default=f"{SQLITEDB}.updated",
    type=str,
)
def update_viewconfs(
    old_hg_name,
    old_site_url,
    old_port,
    old_data_dir,
    new_site_url,
    new_port,
    db_backup_name,
):
    """Update stored viewconfs from one host to another

    The script allows one to update viewconfs saved
    in an existing higlass database. It does so
    by modifying references to tilesets that use
    old-site-url:old-port --> new-site-url:new-port

    old/new-site-urls must include schema (http, https):
    http://localhost
    http://old.host.org
    ...

    if 'old-hg-name' is provided and higlass is running,
    then 'old-site-url,old-port,old-data-dir' are inferred.

    if 'old-hg-name' is NOT provided
    then at least 'old-site-url'and 'old-data-dir'
    are required. This scenario should be invoked
    only when the container is not running, otherwise
    integrity of the DB-backup can not be guaranteed.

    Post 80 is default http port and both
    new-port and old-port defaults to it,
    if not specified otherwise.
    site-url:80 is equivalent to site-url

    Script keeps existing database unchanged,
    but modifies a backed up version located
    in the same path as the original one.

    Running higlass-container would be interrupted
    during database backup.

    """

    # update viewconfs FROM (ORIGIN):
    if old_hg_name is not None:
        # then the container must be running
        try:
            old_site_url = get_site_url(old_hg_name)
            old_port = get_port(old_hg_name)
            old_data_dir = get_data_dir(old_hg_name)
        except docker.errors.NotFound as ex:
            sys.stderr.write(f"Instance not running: {old_hg_name}\n")
    elif (old_site_url is None) or (old_data_dir is None):
        raise ValueError(
            "old-site-url and old-data-dir must be provided,"
            " when instance is not running and no old-hg-name is provided\n"
        )

    # define origin as site_url:port or site_url (when 80)
    origin = old_site_url if (old_port == "80") else f"{old_site_url}:{old_port}"

    # update viewconfs TO (DESTINATION):
    # define destination as site_url:port or site_url (when 80)
    destination = new_site_url if (new_port == "80") else f"{new_site_url}:{new_port}"

    # locate db.sqlite3 and name for the updated version:
    origin_db_path = op.join(old_data_dir, SQLITEDB)
    update_db_path = op.join(old_data_dir, db_backup_name)

    # backup the database using simple copyfile, stop container before
    if old_hg_name is not None:
        _stop([old_hg_name,], False, False, False)
    try:
        shutil.copyfile(origin_db_path, update_db_path)
    except (OSError, IOError):
        sys.stderr.write(f"Failed to copy {origin_db_path} to {update_db_path}")
        sys.exit(-1)
    finally:
        # restart container even if exception occurs:
        if old_hg_name is not None:
            sys.stderr.write(f"Restarting container {old_hg_name} ...\n")
            sys.stderr.flush()
            client = docker.from_env()
            container_name = hg_name_to_container_name(old_hg_name)
            client.containers.get(container_name).restart()
    # alternatively database could be backed up using a "backup"-mechanism:
    # CLI
    #    res = subprocess.run(["sqlite3", origin_db_path, f".backup {update_db_path}"])
    # Python API >= 3.7.0
    #       conn = sqlite3.connect(origin_db_path)
    #       conn.backup(update_db_path)
    # check if it is indeed "atomic" and can be done on a "live" database .

    # now modify the backed-up database "update_db_path" using sqlite3 API:
    conn = None
    try:
        conn = sqlite3.connect(update_db_path)
    except sqlite3.Error as e:
        sys.stderr.write(f"Failed to connect to {update_db_path}\n")
        sys.exit(-1)

    # sql query to update viewconfs, by replacing origin -> destination
    db_query = f"""
            UPDATE tilesets_viewconf
            SET viewconf = replace(viewconf,'{origin}','{destination}')
            """
    # exec sql query
    with conn:
        cur = conn.cursor()
        cur.execute(db_query)

    conn.close()
    # todo: add some stats on - how many viewconfs
    # were updated

    sys.stderr.write(
        f"Backed up version of the database {update_db_path}\n"
        " has been updated and ready for migration\n\n"
        " copy it to the new host along with the media folder\n"
        f" rename the database file back to {SQLITEDB} and restart higlass.\n"
    )
    sys.stderr.flush()
    sys.exit(0)



from setuptools import setup

setup(
    name="higlass-manage",
    author="Peter Kerpedjiev",
    author_email="pkerpedjiev@gmail.com",
    url="https://github.com/pkerpedjiev/higlass-manage",
    description="Wrappers for running the HiGlass Docker container",
    version="0.8.0",
    py_modules=["higlass_manage"],
    packages=["higlass_manage"],
    package_data={"": ["redis/*"]},
    include_package_data=True,
    install_requires=[
        "Click",
        "clodius>=0.10.3",
        "cooler>=0.8.0",
        "pandas>=0.19",
        "docker",
        "requests",
    ],
    entry_points="""
        [console_scripts]
        higlass-manage=higlass_manage.cli:cli
    """,
)



import click
import os.path as op

from higlass_manage.common import get_data_dir


@click.command()
@click.argument("hg_name", nargs=-1)
def logs(hg_name):
    """
    Return the error log for this container
    """
    if len(hg_name) == 0:
        hg_name = "default"
    else:
        hg_name = hg_name[0]

    data_dir = get_data_dir(hg_name)
    log_location = op.join(data_dir, "log", "hgs.log")

    with open(log_location, "r") as f:
        for line in f:
            print(line, end="")



import click
import clodius.cli.aggregate as cca
import ntpath
import os.path as op
import sys

from higlass_manage.common import fill_filetype_and_datatype
from higlass_manage.common import import_file
from higlass_manage.common import get_temp_dir
from higlass_manage.start import _start


@click.command()
@click.argument("filename")
@click.option(
    "--hg-name",
    default="default",
    help="The name of the higlass container to import this file to",
)
@click.option(
    "--filetype", default=None, help="The type of file to ingest (e.g. cooler)"
)
@click.option(
    "--datatype", default=None, help="The data type of in the input file (e.g. matrix)"
)
@click.option(
    "--assembly", default=None, help="The assembly that this data is mapped to"
)
@click.option("--name", default=None, help="The name to use for this file")
@click.option("--uid", default=None, help="The uuid to use for this file")
@click.option(
    "--no-upload",
    default=None,
    is_flag=True,
    help="Do not copy the file to the media directory. File must already be in the media directory.",
)
@click.option(
    "--chromsizes-filename",
    default=None,
    help="A set of chromosome sizes to use for bed and bedpe files",
)
@click.option(
    "--has-header",
    default=False,
    is_flag=True,
    help="Does the input file have column header information (only relevant for bed or bedpe files)",
)
@click.option(
    "--project-name",
    default=None,
    help="Group this tileset with others by specifying a project name",
)
def ingest(
    filename,
    hg_name,
    filetype=None,
    datatype=None,
    assembly=None,
    name=None,
    chromsizes_filename=None,
    has_header=False,
    uid=None,
    no_upload=None,
    project_name=None,
):
    """
    Ingest a dataset
    """
    _ingest(
        filename,
        hg_name,
        filetype,
        datatype,
        assembly,
        name,
        chromsizes_filename,
        has_header,
        uid,
        no_upload,
        project_name,
    )


def _ingest(
    filename,
    hg_name,
    filetype=None,
    datatype=None,
    assembly=None,
    name=None,
    chromsizes_filename=None,
    has_header=False,
    uid=None,
    no_upload=None,
    project_name=None,
):

    try:
        get_temp_dir(hg_name)
    except Exception:
        print("HiGlass not running. Starting...")
        _start(hg_name=hg_name)

    if not no_upload and (not op.exists(filename) and not op.islink(filename)):
        print("File not found:", filename, file=sys.stderr)
        return None

    # guess filetype and datatype if they're None
    (filetype, datatype) = fill_filetype_and_datatype(filename, filetype, datatype)

    temp_dir = get_temp_dir(hg_name)
    (to_import, filetype) = aggregate_file(
        filename,
        filetype,
        assembly,
        chromsizes_filename,
        has_header,
        no_upload,
        temp_dir,
    )

    return import_file(
        hg_name,
        to_import,
        filetype,
        datatype,
        assembly,
        name,
        uid,
        no_upload,
        project_name,
    )


def aggregate_file(
    filename, filetype, assembly, chromsizes_filename, has_header, no_upload, tmp_dir
):
    if filetype == "bedfile":
        if no_upload:
            raise Exception(
                "Bedfile files need to be aggregated and cannot be linked. Consider not using the --link-file option",
                file=sys.stderr,
            )

        if assembly is None and chromsizes_filename is None:
            print(
                "An assembly or set of chromosome sizes is required when importing bed files. Please specify one or the other using the --assembly or --chromsizes-filename parameters",
                file=sys.stderr,
            )
            return

        output_file = op.join(tmp_dir, ntpath.basename(filename) + ".beddb")

        print("Aggregating bedfile")
        cca._bedfile(
            filename,
            output_file,
            assembly,
            importance_column="random",
            has_header=has_header,
            chromosome=None,
            max_per_tile=50,
            delimiter=None,
            chromsizes_filename=chromsizes_filename,
            offset=0,
            tile_size=1024,
        )

        to_import = output_file

        # because we aggregated the file, the new filetype is beddb
        filetype = "beddb"
        return (to_import, filetype)
    elif filetype == "bedpe":
        if no_upload:
            raise Exception(
                "Bedpe files need to be aggregated and cannot be linked. Consider not using the --link-file option",
                file=sys.stderr,
            )
        if assembly is None and chromsizes_filename is None:
            print(
                "An assembly or set of chromosome sizes is required when importing bed files. Please specify one or the other using the --assembly or --chromsizes-filename parameters",
                file=sys.stderr,
            )
            return

        output_file = op.join(tmp_dir, filename + ".bed2ddb")

        print("Aggregating bedpe (output_file: {}".format(output_file))
        cca._bedpe(
            filename,
            output_file,
            assembly,
            importance_column="random",
            has_header=has_header,
            chromosome=None,
            max_per_tile=50,
            chromsizes_filename=chromsizes_filename,
            tile_size=1024,
        )

        to_import = output_file

        # because we aggregated the file, the new filetype is beddb
        filetype = "bed2ddb"
        return (to_import, filetype)
    else:
        return (filename, filetype)
